@@ -1,4 +1,4 @@
-# Copyright (c) 2009 CentralNic Ltd. All rights reserved. This program is
+# Copyright (c) 2010 CentralNic Ltd. All rights reserved. This program is
 # free software; you can redistribute it and/or modify it under the same
 # terms as Perl itself.
 # 
@@ -7,25 +7,26 @@ package Net::EPP::Simple;
 use Carp;
 use Digest::SHA1 qw(sha1_hex);
 use Net::EPP::Frame;
+use Net::EPP::ResponseCodes;
 use Time::HiRes qw(time);
 use UNIVERSAL qw(isa);
 use base qw(Net::EPP::Client);
 use constant EPP_XMLNS	=> 'urn:ietf:params:xml:ns:epp-1.0';
 use vars qw($Error $Code $Message);
-use warnings;
 use strict;
+use warnings;
 
 our $Error	= '';
-our $Code	= 1000;
+our $Code	= OK;
 our $Message	= '';
 
 =pod
 
-=head1 NAME
+=head1 Name
 
 Net::EPP::Simple - a simple EPP client interface for the most common jobs
 
-=head1 SYNOPSIS
+=head1 Synopsis
 
 	#!/usr/bin/perl
 	use Net::EPP::Simple;
@@ -48,7 +49,7 @@ Net::EPP::Simple - a simple EPP client interface for the most common jobs
 
 	}
 
-=head1 DESCRIPTION
+=head1 Description
 
 EPP is the Extensible Provisioning Protocol. EPP (defined in RFC 4930) is an
 application layer client-server protocol for the provisioning and management of
@@ -65,7 +66,7 @@ response frames behind a simple, Perlish interface.
 It is based on the C<Net::EPP::Client> module and uses C<Net::EPP::Frame>
 to build request frames.
 
-=head1 CONSTRUCTOR
+=head1 Constructor
 
 The constructor for C<Net::EPP::Simple> has the same general form as the
 one for C<Net::EPP::Client>, but with the following exceptions:
@@ -80,41 +81,212 @@ one for C<Net::EPP::Client>, but with the following exceptions:
 
 =item * The C<timeout> parameter controls how long the client waits for a response from the server before returning an error.
 
+=item * if C<debug> is set, C<Net::EPP::Simple> will output verbose debugging information on C<STDERR>, including all frames sent to and received from the server.
+
+=item * C<reconnect> can be used to disable automatic reconnection (it is enabled by default). Before sending a frame to the server, C<Net::EPP::Simple> will send a C<E<lt>helloE<gt>> to check that the connection is up, if not, it will try to reconnect, aborting after the I<n>th time, where I<n> is the value of C<reconnect> (the default is 3).
+
 =back
 
 The constructor will establish a connection to the server and retrieve the
-greeting (which is available via C<$epp-E<gt>{greeting}>) and then send a
-C<E<lt>loginE<gt>> request.
+greeting (which is available via $epp-E<gt>{greeting}) and then send a
+E<lt>loginE<gt> request.
 
 If the login fails, the constructor will return C<undef> and set
 C<$Net::EPP::Simple::Error> and C<$Net::EPP::Simple::Code>.
 
+=head2 Client and Server SSL options
+
+RFC 5730 requires that all EPP instances must be protected using "mutual,
+strong client-server authentication". In practice, this means that both
+client and server must present an SSL certificate, and that they must
+both verify the certificate of their peer.
+
+=head3 Server Certificate Verification
+
+C<Net::EPP::Simple> will verify the certificate presented by a server if
+the C<verify>, and either C<ca_file> or C<ca_path> are passed to the
+constructor:
+
+	my $epp = Net::EPP::Simple->new(
+		host	=> 'epp.nic.tld',
+		user	=> 'my-id',
+		pass	=> 'my-password',
+		verify	=> 1,
+		ca_file	=> '/etc/pki/tls/certs/ca-bundle.crt',
+		ca_path	=> '/etc/pki/tls/certs',
+	);
+
+C<Net::EPP::Simple> will fail to connect to the server if the
+certificate is not valid.
+
+=head3 SSL Cipher Selection
+
+You can restrict the ciphers that you will use to connect to the server
+by passing a C<ciphers> parameter to the constructor. This is a colon-
+separated list of cipher names and aliases. See L<http://www.openssl.org/docs/apps/ciphers.html#CIPHER_STRINGS>
+for further details. As an example, the following cipher list is
+suggested for clients who wish to ensure high-security connections to
+servers:
+
+	HIGH:!ADH:!MEDIUM:!LOW:!SSLv2:!EXP
+
+=head3 Client Certificates
+
+If you are connecting to an EPP server which requires a client
+certificate, you can configure C<Net::EPP::Simple> to use one as
+follows:
+
+	my $epp = Net::EPP::Simple->new(
+		host		=> 'epp.nic.tld',
+		user		=> 'my-id',
+		pass		=> 'my-password',
+		key		=> '/path/to/my.key',
+		cert		=> '/path/to/my.crt',
+		passphrase	=> 'foobar123',
+	);
+
+C<key> is the filename of the private key, C<cert> is the filename of
+the certificate. If the private key is encrypted, the C<passphrase>
+parameter will be used to decrypt it.
+
+=head2 Configuration File
+
+C<Net::EPP::Simple> supports the use of a simple configuration file. To 
+use this feature, you need to install the L<Config::Simple> module.
+
+When starting up, C<Net::EPP::Simple> will look for 
+C<$HOME/.net-epp-simple-rc>. This file is an ini-style configuration 
+file.
+
+=head3 Default Options
+
+You can specify default options for all EPP servers using the C<[default]>
+section:
+
+	[default]
+	default=epp.nic.tld
+	debug=1
+
+=head3 Server Specific Options
+
+You can specify options for for specific EPP servers by giving each EPP server
+its own section:
+
+	[epp.nic.tld]
+	user=abc123
+	pass=foo2bar
+	port=777
+	ssl=0
+
+This means that when you write a script that uses C<Net::EPP::Simple>, you can
+do the following:
+
+	# config file has a default server:
+	my $epp = Net::EPP::Simple->new;
+
+	# config file has connection options for this EPP server:
+	my $epp = Net::EPP:Simple->new('host' => 'epp.nic.tld');
+
+Any parameters provided to the constructor will override those in the config
+file.
+
 =cut
 
 sub new {
 	my ($package, %params) = @_;
 	$params{dom}		= 1;
+
+	eval 'use Config::Simple';
+	if (!$@) {
+		# we have Config::Simple, so let's try to parse the RC file:
+		my $rcfile = $ENV{'HOME'}.'/.net-epp-simple-rc';
+		if (-e $rcfile) {
+			my $config = Config::Simple->new($rcfile);
+
+			# if no host was defined in the constructor, use the default (if specified):
+			if (!defined($params{'host'}) && $config->param('default.default')) {
+				$params{'host'} = $config->param('default.default');
+			}
+
+			# if no debug level was defined in the constructor, use the default (if specified):
+			if (!defined($params{'debug'}) && $config->param('default.debug')) {
+				$params{'debug'} = $config->param('default.debug');
+			}
+
+			# grep through the file's values for settings for the selected host:
+			my %vars = $config->vars;
+			foreach my $key (grep { /^$params{'host'}\./ } keys(%vars)) {
+				my $value = $vars{$key};
+				$key =~ s/^$params{'host'}\.//;
+				$params{$key} = $value unless (defined($params{$key}));
+			}
+		}
+	}
+
 	$params{port}		= (defined($params{port}) && int($params{port}) > 0 ? $params{port} : 700);
 	$params{ssl}		= ($params{no_ssl} ? undef : 1);
 
 	my $self = $package->SUPER::new(%params);
 
+	$self->{user}		= $params{user};
+	$self->{pass}		= $params{pass};
 	$self->{debug} 		= (defined($params{debug}) ? int($params{debug}) : undef);
 	$self->{timeout}	= (defined($params{timeout}) && int($params{timeout}) > 0 ? $params{timeout} : 5);
+	$self->{reconnect}	= (defined($params{reconnect}) ? int($params{reconnect}) : 3);
 	$self->{connected}	= undef;
 	$self->{authenticated}	= undef;
+	$self->{connect}	= (exists($params{connect}) ? $params{connect} : 1);
+	$self->{login}		= (exists($params{login}) ? $params{login} : 1);
+	$self->{key}		= $params{key};
+	$self->{cert}		= $params{cert};
+	$self->{key_passphrase}	= $params{key_passphrase};
+	$self->{verify}		= $params{verify};
+	$self->{ca_file}	= $params{ca_file};
+	$self->{ca_path}	= $params{ca_path};
+	$self->{ciphers}	= $params{ciphers};
 
 	bless($self, $package);
 
+	if ($self->{connect}) {
+		return ($self->_connect($self->{login}) ? $self : undef);
+
+	} else {
+		return $self;
+
+	}
+}
+
+sub _connect {
+	my ($self, $login) = @_;
+
+	my %params;
+
+	$params{SSL_cipher_list} = $self->{ciphers} if (defined($self->{ssl}) && defined($self->{ciphers}));
+
+	if (defined($self->{key}) && defined($self->{cert}) && defined($self->{ssl})) {
+		$self->debug('configuring client certificate parameters');
+		$params{SSL_key_file}	= $self->{key};
+		$params{SSL_cert_file}	= $self->{cert};
+		$params{SSL_passwd_cb}	= sub { $self->{key_passphrase} };
+	}
+
+	if (defined($self->{ssl}) && defined($self->{verify})) {
+		$self->debug('configuring server verification');
+		$params{SSL_verify_mode}	= 0x01;
+		$params{SSL_ca_file}		= $self->{ca_file};
+		$params{SSL_ca_path}		= $self->{ca_path};
+	}
+
 	$self->debug(sprintf('Attempting to connect to %s:%d', $self->{host}, $self->{port}));
 	eval {
-		$self->{greeting} = $self->connect;
+		$self->{greeting} = $self->connect(%params);
 	};
 	if ($@ ne '' || ref($self->{greeting}) ne 'Net::EPP::Frame::Response') {
 		chomp($@);
 		$@ =~ s/ at .+ line .+$//;
-		$Code = 2400;
-		$Message = $@;
+		$self->debug($@);
+		$Code = COMMAND_FAILED;
+		$Error = $Message = $@;
 		return undef;
 	}
 
@@ -122,12 +294,27 @@ sub new {
 
 	map { $self->debug('S: '.$_) } split(/\n/, $self->{greeting}->toString(1));
 
-	$self->debug('Connected OK, preparing login frame');
+	$self->debug('Connected OK');
 
+	if ($login) {
+		return $self->_login;
+
+	} else {
+		return 1;
+
+	}
+}
+
+sub _login {
+	my $self = shift;
+
+	$self->debug('preparing login frame');
 	my $login = Net::EPP::Frame::Command::Login->new;
 
-	$login->clID->appendText($params{user});
-	$login->pw->appendText($params{pass});
+	$login->clID->appendText($self->{user});
+	$login->pw->appendText($self->{pass});
+	$login->version->appendText($self->{greeting}->getElementsByTagNameNS(EPP_XMLNS, 'version')->shift->firstChild->data);
+	$login->lang->appendText($self->{greeting}->getElementsByTagNameNS(EPP_XMLNS, 'lang')->shift->firstChild->data);
 
 	my $objects = $self->{greeting}->getElementsByTagNameNS(EPP_XMLNS, 'objURI');
 	while (my $object = $objects->shift) {
@@ -136,13 +323,18 @@ sub new {
 		$login->svcs->appendChild($el);
 	}
 	$objects = $self->{greeting}->getElementsByTagNameNS(EPP_XMLNS, 'extURI');
+	my $svcext;
+	if ($objects->size) {
+		$svcext = $login->createElement('svcExtension');
+		$login->svcs->appendChild($svcext);
+	}
 	while (my $object = $objects->shift) {
-		my $el = $login->createElement('objURI');
+		my $el = $login->createElement('extURI');
 		$el->appendText($object->firstChild->data);
-		$login->svcs->appendChild($el);
+		$svcext->appendChild($el);
 	}
 
-	$self->debug(sprintf("Attempting to login as client ID '%s'", $params{user}));
+	$self->debug(sprintf("Attempting to login as client ID '%s'", $self->{user}));
 	my $response = $self->request($login);
 
 	$Code = $self->_get_response_code($response);
@@ -159,7 +351,7 @@ sub new {
 
 	}
 
-	return $self;
+	return 1;
 }
 
 =pod
@@ -222,27 +414,32 @@ sub _check {
 		return undef;
 	}
 
-	my $response = $self->request($frame);
+	my $response = $self->_request($frame);
 
-	$Code = $self->_get_response_code($response);
-	$Message = $self->_get_message($response);
-
-	if ($Code > 1999) {
-		$Error = sprintf("Server returned a %d code", $Code);
+	if (!$response) {
 		return undef;
 
 	} else {
-		my $xmlns = (Net::EPP::Frame::ObjectSpec->spec($type))[1];
-		my $key;
-		if ($type eq 'domain' || $type eq 'host') {
-			$key = 'name';
+		$Code = $self->_get_response_code($response);
+		$Message = $self->_get_message($response);
 
-		} elsif ($type eq 'contact') {
-			$key = 'id';
+		if ($Code > 1999) {
+			$Error = sprintf("Server returned a %d code", $Code);
+			return undef;
+
+		} else {
+			my $xmlns = (Net::EPP::Frame::ObjectSpec->spec($type))[1];
+			my $key;
+			if ($type eq 'domain' || $type eq 'host') {
+				$key = 'name';
+
+			} elsif ($type eq 'contact') {
+				$key = 'id';
+
+			}
+			return $response->getNode($xmlns, $key)->getAttribute('avail');
 
 		}
-		return $response->getNode($xmlns, $key)->getAttribute('avail');
-
 	}
 }
 
@@ -252,11 +449,11 @@ sub _check {
 
 You can retrieve information about an object by using one of the following:
 
-	my $info = $epp->domain_info($domain);
+	my $info = $epp->domain_info($domain, $authInfo, $follow);
 
 	my $info = $epp->host_info($host);
 
-	my $info = $epp->contact_info($contact);
+	my $info = $epp->contact_info($contact, $authInfo);
 
 C<Net::EPP::Simple> will construct an C<E<lt>infoE<gt>> frame and send
 it to the server, then parse the response into a simple hash ref. The
@@ -264,11 +461,49 @@ layout of the hash ref depends on the object in question. If there is an
 error, these methods will return C<undef>, and you can then check
 C<$Net::EPP::Simple::Error> and C<$Net::EPP::Simple::Code>.
 
+If C<$authInfo> is defined, it will be sent to the server as per RFC
+5731, Section 3.1.2 and RFC 5733, Section 3.1.2. If the supplied
+authInfo code is validated by the registry, additional information will
+appear in the response. If it is invalid, you should get an error.
+
+If the C<$follow> parameter is true, then C<Net::EPP::Simple> will also
+retrieve the relevant host and contact details for a domain: instead of
+returning an object name or ID for the domain's registrant, contact
+associations, DNS servers or subordinate hosts, the values will be
+replaced with the return value from the appropriate C<host_info()> or
+C<contact_info()> command (unless there was an error, in which case the
+original object ID will be used instead).
+
 =cut
 
 sub domain_info {
-	my ($self, $domain) = @_;
-	return $self->_info('domain', $domain);
+	my ($self, $domain, $authInfo, $follow) = @_;
+	my $result = $self->_info('domain', $domain, $authInfo);
+	return $result if (ref($result) ne 'HASH' || !$follow);
+
+	if (defined($result->{'ns'}) && ref($result->{'ns'}) eq 'ARRAY') {
+		for (my $i = 0 ; $i < scalar(@{$result->{'ns'}}) ; $i++) {
+			my $info = $self->host_info($result->{'ns'}->[$i]);
+			$result->{'ns'}->[$i] = $info if (ref($info) eq 'HASH');
+		}
+	}
+
+	if (defined($result->{'hosts'}) && ref($result->{'hosts'}) eq 'ARRAY') {
+		for (my $i = 0 ; $i < scalar(@{$result->{'hosts'}}) ; $i++) {
+			my $info = $self->host_info($result->{'hosts'}->[$i]);
+			$result->{'hosts'}->[$i] = $info if (ref($info) eq 'HASH');
+		}
+	}
+
+	my $info = $self->contact_info($result->{'registrant'});
+	$result->{'registrant'} = $info if (ref($info) eq 'HASH');
+
+	foreach my $type (keys(%{$result->{'contacts'}})) {
+		my $info = $self->contact_info($result->{'contacts'}->{$type});
+		$result->{'contacts'}->{$type} = $info if (ref($info) eq 'HASH');
+	}
+
+	return $result;
 }
 
 sub host_info {
@@ -277,12 +512,12 @@ sub host_info {
 }
 
 sub contact_info {
-	my ($self, $contact) = @_;
-	return $self->_info('contact', $contact);
+	my ($self, $contact, $authInfo) = @_;
+	return $self->_info('contact', $contact, $authInfo);
 }
 
 sub _info {
-	my ($self, $type, $identifier) = @_;
+	my ($self, $type, $identifier, $authInfo) = @_;
 	my $frame;
 	if ($type eq 'domain') {
 		$frame = Net::EPP::Frame::Command::Info::Domain->new;
@@ -299,29 +534,44 @@ sub _info {
 	} else {
 		$Error = "Unknown object type '$type'";
 		return undef;
+
 	}
 
-	my $response = $self->request($frame);
+	if (defined($authInfo) && $authInfo ne '') {
+		$self->debug('adding authInfo element to request frame');
+		my $el = $frame->createElement((Net::EPP::Frame::ObjectSpec->spec($type))[0].':authInfo');
+		my $pw = $frame->createElement((Net::EPP::Frame::ObjectSpec->spec($type))[0].':pw');
+		$pw->appendChild($frame->createTextNode($authInfo));
+		$el->appendChild($pw);
+		$frame->getNode((Net::EPP::Frame::ObjectSpec->spec($type))[1], 'info')->appendChild($el);
+	}
 
-	$Code = $self->_get_response_code($response);
-	$Message = $self->_get_message($response);
+	my $response = $self->_request($frame);
 
-	if ($Code > 1999) {
-		$Error = sprintf("Server returned a %d code", $Code);
+	if (!$response) {
 		return undef;
 
 	} else {
-		my $infData = $response->getNode((Net::EPP::Frame::ObjectSpec->spec($type))[1], 'infData');
+		$Code = $self->_get_response_code($response);
+		$Message = $self->_get_message($response);
 
-		if ($type eq 'domain') {
-			return $self->_domain_infData_to_hash($infData);
+		if ($Code > 1999) {
+			$Error = sprintf("Server returned a %d code", $Code);
+			return undef;
 
-		} elsif ($type eq 'contact') {
-			return $self->_contact_infData_to_hash($infData);
+		} else {
+			my $infData = $response->getNode((Net::EPP::Frame::ObjectSpec->spec($type))[1], 'infData');
 
-		} elsif ($type eq 'host') {
-			return $self->_host_infData_to_hash($infData);
+			if ($type eq 'domain') {
+				return $self->_domain_infData_to_hash($infData);
 
+			} elsif ($type eq 'contact') {
+				return $self->_contact_infData_to_hash($infData);
+
+			} elsif ($type eq 'host') {
+				return $self->_host_infData_to_hash($infData);
+
+			}
 		}
 	}
 }
@@ -568,10 +818,12 @@ sub _contact_infData_to_hash {
 
 	foreach my $name ('voice', 'fax') {
 		my $els = $infData->getElementsByLocalName($name);
-		if ($els->size == 1) {
+		if (defined($els) && $els->size == 1) {
 			my $el = $els->shift;
-			$hash->{$name} = $el->textContent;
-			$hash->{$name} .= 'x'.$el->getAttribute('x') if ($el->getAttribute('x') ne '');
+			if (defined($el)) {
+				$hash->{$name} = $el->textContent;
+				$hash->{$name} .= 'x'.$el->getAttribute('x') if (defined($el->getAttribute('x')) && $el->getAttribute('x') ne '');
+			}
 		}
 	}
 
@@ -660,7 +912,7 @@ sub _transfer_request {
 	eval("\$frame = $class->new");
 	if ($@ || ref($frame) ne $class) {
 		$Error = "Error building request frame: $@";
-		$Code = 2400;
+		$Code = COMMAND_FAILED;
 		return undef;
 
 	} else {
@@ -680,27 +932,33 @@ sub _transfer_request {
 		$frame->setPeriod(int($period)) if ($op eq 'request');
 	}
 
-	my $response = $self->request($frame);
+	my $response = $self->_request($frame);
 
-	$Code = $self->_get_response_code($response);
-	$Message = $self->_get_message($response);
 
-	if ($Code > 1999) {
-		$Error = $response->msg;
+	if (!$response) {
 		return undef;
 
-	} elsif ($op eq 'query' || $op eq 'request') {
-		my $trnData = $response->getElementsByLocalName('trnData')->shift;
-		my $hash = {};
-		foreach my $child ($trnData->childNodes) {
-			$hash->{$child->localName} = $child->textContent;
-		}
-
-		return $hash;
-
 	} else {
-		return 1;
+		$Code = $self->_get_response_code($response);
+		$Message = $self->_get_message($response);
 
+		if ($Code > 1999) {
+			$Error = $response->msg;
+			return undef;
+
+		} elsif ($op eq 'query' || $op eq 'request') {
+			my $trnData = $response->getElementsByLocalName('trnData')->shift;
+			my $hash = {};
+			foreach my $child ($trnData->childNodes) {
+				$hash->{$child->localName} = $child->textContent;
+			}
+
+			return $hash;
+
+		} else {
+			return 1;
+
+		}
 	}
 }
 
@@ -796,14 +1054,44 @@ When creating a new domain object, you may also specify a C<period> key, like so
 
 	$epp->create_domain($domain);
 
-C<Net::EPP::Simple> assumes the registry uses the host object model rather
-than the host attribute model.
+The C<period> key is assumed to be in years rather than months. C<Net::EPP::Simple>
+assumes the registry uses the host object model rather than the host attribute model.
 
 =cut
 
 sub create_domain {
 	my ($self, $domain) = @_;
-	croak("Unfinished method create_domain()");
+
+	print Data::Dumper::Dumper($domain);
+
+	my $frame = Net::EPP::Frame::Command::Create::Domain->new;
+	$frame->setDomain($domain->{'name'});
+	$frame->setPeriod($domain->{'period'});
+	$frame->setRegistrant($domain->{'registrant'});
+	$frame->setContacts($domain->{'contacts'});
+	$frame->setNS(@{$domain->{'ns'}}) if $domain->{'ns'} and @{$domain->{'ns'}};
+
+	$frame->setAuthInfo($domain->{authInfo}) if ($domain->{authInfo} ne '');
+
+	my $response = $self->_request($frame);
+
+
+	if (!$response) {
+		return undef;
+
+	} else {
+		$Code = $self->_get_response_code($response);
+		$Message = $self->_get_message($response);
+
+		if ($Code > 1999) {
+			$Error = $response->msg;
+			return undef;
+
+		} else {
+			return 1;
+
+		}
+	}
 }
 
 sub create_host {
@@ -816,10 +1104,6 @@ sub create_contact {
 	my $frame = Net::EPP::Frame::Command::Create::Contact->new;
 
 	$frame->setContact($contact->{id});
-	$frame->setEmail($contact->{email});
-	$frame->setVoice($contact->{voice}) if ($contact->{voice} ne '');
-	$frame->setFax($contact->{fax}) if ($contact->{fax} ne '');
-	$frame->setAuthInfo($contact->{authInfo}) if ($contact->{authInfo} ne '');
 
 	if (ref($contact->{postalInfo}) eq 'HASH') {
 		foreach my $type (keys(%{$contact->{postalInfo}})) {
@@ -832,26 +1116,35 @@ sub create_contact {
 		}
 	}
 
+	$frame->setVoice($contact->{voice}) if ($contact->{voice} ne '');
+	$frame->setFax($contact->{fax}) if ($contact->{fax} ne '');
+	$frame->setEmail($contact->{email});
+	$frame->setAuthInfo($contact->{authInfo}) if ($contact->{authInfo} ne '');
+
 	if (ref($contact->{status}) eq 'ARRAY') {
 		foreach my $status (grep { /^client/ } @{$contact->{status}}) {
 			$frame->appendStatus($status);
 		}
 	}
 
-	my $response = $self->request($frame);
+	my $response = $self->_request($frame);
 
-	$Code = $self->_get_response_code($response);
-	$Message = $self->_get_message($response);
-
-	if ($Code > 1999) {
-		$Error = $response->msg;
+	if (!$response) {
 		return undef;
 
 	} else {
-		return 1;
+		$Code = $self->_get_response_code($response);
+		$Message = $self->_get_message($response);
 
+		if ($Code > 1999) {
+			$Error = $response->msg;
+			return undef;
+
+		} else {
+			return 1;
+
+		}
 	}
-
 }
 
 sub update_domain {
@@ -929,18 +1222,24 @@ sub _delete {
 
 	}
 
-	my $response = $self->request($frame);
+	my $response = $self->_request($frame);
 
-	$Code = $self->_get_response_code($response);
-	$Message = $self->_get_message($response);
 
-	if ($Code > 1999) {
-		$Error = sprintf("Server returned a %d code", $Code);
+	if (!$response) {
 		return undef;
 
 	} else {
-		return 1;
+		$Code = $self->_get_response_code($response);
+		$Message = $self->_get_message($response);
 
+		if ($Code > 1999) {
+			$Error = sprintf("Server returned a %d code", $Code);
+			return undef;
+
+		} else {
+			return 1;
+
+		}
 	}
 }
 
@@ -966,6 +1265,46 @@ Returns the a C<Net::EPP::Frame::Greeting> object representing the greeting retu
 
 sub greeting { $_[0]->{greeting} }
 
+sub ping {
+	my $self = shift;
+	my $hello = Net::EPP::Frame::Hello->new;
+	my $response = $self->request($hello);
+
+	return (isa($response, 'XML::LibXML::Document') ? 1 : undef);
+}
+
+sub _request {
+	my ($self, $frame) = @_;
+
+	if ($self->{reconnect} > 0) {
+		if (!$self->ping) {
+			$self->debug('connection seems dead, trying to reconnect');
+			for (1..$self->{reconnect}) {
+				$self->debug("attempt #$_");
+				if ($self->_connect) {
+					$self->debug("attempt #$_ succeeded");
+					return $self->request($frame);
+
+				} else {
+					$self->debug("attempt #$_ failed, sleeping");
+					sleep($self->{timeout});
+
+				}
+			}
+			$self->debug('unable to reconnect!');
+			return undef;
+
+		} else {
+			return $self->request($frame);
+
+		}
+
+	} else {
+		return $self->request($frame);
+
+	}
+}
+
 =pod
 
 =head1 Overridden Methods From C<Net::EPP::Client>
@@ -983,6 +1322,13 @@ frame back up to C<Net::EPP::Client>.
 
 sub request {
 	my ($self, $frame) = @_;
+	# Make sure we start with blank variables
+	$Code		= undef;
+	$Error		= '';
+	$Message	= '';
+
+	$frame->clTRID->appendText(sha1_hex(ref($self).time().$$)) if (isa($frame, 'Net::EPP::Frame::Command'));
+
 	$self->debug(sprintf('sending a %s to the server', ref($frame)));
 	if (isa($frame, 'XML::LibXML::Document')) {
 		map { $self->debug('C: '.$_) } split(/\n/, $frame->toString(1));
@@ -991,11 +1337,11 @@ sub request {
 		map { $self->debug('C: '.$_) } split(/\n/, $frame);
 
 	}
-	$frame->clTRID->appendText(sha1_hex(ref($self).time().$$)) if (isa($frame, 'XML::LibXML::Node'));
+
 	my $response = $self->SUPER::request($frame);
-	if (isa($response, 'XML::LibXML::Document')) {
-		map { $self->debug('S: '.$_) } split(/\n/, $response->toString(1));
-	}
+
+	map { $self->debug('S: '.$_) } split(/\n/, $response->toString(1)) if (isa($response, 'XML::LibXML::Document'));
+
 	return $response;
 }
 
@@ -1011,19 +1357,20 @@ network errors. If such an error occurs it will return C<undef>.
 sub get_frame {
 	my $self = shift;
 	my $frame;
-	$self->debug(sprintf('transmitting frame, waiting %d seconds before timeout', $self->{timeout}));
+	$self->debug(sprintf('reading frame, waiting %d seconds before timeout', $self->{timeout}));
 	eval {
 		local $SIG{ALRM} = sub { die "alarm\n" };
-		$self->debug('setting alarm');
+		$self->debug('setting timeout alarm for receiving frame');
 		alarm($self->{timeout});
 		$frame = $self->SUPER::get_frame();
-		$self->debug('unsetting alarm');
+		$self->debug('unsetting timeout alarm after successful receive');
 		alarm(0);
 	};
 	if ($@ ne '') {
-		$self->debug('unsetting alarm');
+		$self->debug('unsetting timeout alarm after alarm was triggered');
 		alarm(0);
-		$Error = "get_frame() timed out\n";
+		$Code = COMMAND_FAILED;
+		$Error = $Message = "get_frame() timed out\n";
 		return undef;
 
 	} else {
@@ -1065,6 +1412,7 @@ sub logout {
 	}
 	$self->debug('disconnecting from server');
 	$self->disconnect;
+	$self->{connected} = 0;
 	return 1;
 }
 
@@ -1081,7 +1429,7 @@ sub debug {
 
 =pod
 
-=head1 PACKAGE VARIABLES
+=head1 Package Variables
 
 =head2 $Net::EPP::Simple::Error
 
@@ -1105,13 +1453,13 @@ parameters being passed to a method, or a network error) then this will
 be set to 2400 (C<COMMAND_FAILED>). See L<Net::EPP::ResponseCodes> for
 more information about thes codes.
 
-=head1 AUTHOR
+=head1 Author
 
 CentralNic Ltd (L<http://www.centralnic.com/>).
 
-=head1 COPYRIGHT
+=head1 Copyright
 
-This module is (c) 2007 CentralNic Ltd. This module is free software; you can
+This module is (c) 2011 CentralNic Ltd. This module is free software; you can
 redistribute it and/or modify it under the same terms as Perl itself.
 
 =head1 SEE ALSO
@@ -1124,9 +1472,9 @@ redistribute it and/or modify it under the same terms as Perl itself.
 
 =item * L<Net::EPP::Proxy>
 
-=item * RFCs 4930 and RFC 4934, available from L<http://www.ietf.org/>.
+=item * RFCs 5730 and RFC 4934, available from L<http://www.ietf.org/>.
 
-=item * The CentralNic EPP site at L<http://www.centralnic.com/resellers/epp>.
+=item * The CentralNic EPP site at L<http://www.centralnic.com/registrars/epp>.
 
 =back
 
